"""
This module provides package information for hbutils.model.

It defines basic metadata about the hbutils.model package including its name,
homepage URL, and reStructuredText formatted reference string.

:var _PACKAGE_NAME: The name of the hbutils.model package.
:vartype _PACKAGE_NAME: str

:var _PACKAGE_HOMEPAGE: The URL to the hbutils.model package documentation homepage.
:vartype _PACKAGE_HOMEPAGE: str

:var _PACKAGE_RST: A reStructuredText formatted string containing a link to the package homepage.
:vartype _PACKAGE_RST: str
"""

# hbutils.model's package info
_PACKAGE_NAME = 'hbutils.model'
_PACKAGE_HOMEPAGE = 'https://hansbug.github.io/hbutils/main/api_doc/model/index.html'
_PACKAGE_RST = f'`{_PACKAGE_NAME} <{_PACKAGE_HOMEPAGE}>`_'
