"""
Overview:
    Meta information for hbutils package.

This module contains metadata constants for the hbutils package, including version,
author information, and project description. These constants are typically used in
setup configuration and package distribution.
"""

#: Title of this project (should be `hbutils`).
__TITLE__ = "hbutils"

#: Version of this project.
__VERSION__ = "0.12.0"

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'Some useful functions and classes in Python infrastructure development.'

#: Author of this project.
__AUTHOR__ = "HansBug"

#: Email of the authors'.
__AUTHOR_EMAIL__ = "hansbug@buaa.edu.cn"
